/*
  arrays.pl
  
  @author Francois Fages
  @email Francois.Fages@inria.fr
  @license LGPL-2

  @version 1.1.3

  
  Multidimensional arrays with Array[Indices] functional notation and list conversions.

  The indices are evaluated and can be expressions containing shorthands.

*/

:- module(
	  arrays,
	  [
	   array/1,
	   array/2,
	   
	   cell/3,
	   cell/2,
	   op(100, yf, []),
	   
	   array_lists/2,
	   array_list/2,
	   
	   set_cell/3,
	   set_cell/2,
	   nb_set_cell/3,
	   nb_set_cell/2
	   ]
	 ).

/** <module> multidimensional arrays with conversions to lists and Array[Indices] functional notation.

@author Francois Fages
@version 1.1.3

This module provides an implementation of multidimensional arrays by terms.

The array indices are integers starting at 1 and the dimension of an array is a list of integers.

array_list/2 (resp. array_lists/2) makes conversions between an array and a list (resp. of lists for multi-dimensional arrays), which can be used to initialize an array to a list of values.

==
?- array_lists(A, [[1, 2, 3], [4, 5, 6]]), array(A, Dim).
A = array(array(1, 2, 3), array(4, 5, 6)),
Dim = [2, 3].
==

  Array cells are accessed by unification with predicate cell/3.

This module includes module comprehension.pl for bounded quantification and is compatible with attributed variables, 
clpfd and clpr libraries for creating arrays of constrained variables, and posting constraints on subscripted variables.

Array cells can also be modified by destructive assignment, backtrackable or not, with set_cell/3 and nb_set_cell/3.

==
?- array(A, [3]), cell(A, [2], v).
A = array(_, v, _).

?- array(A, [2, 3]), cell(A, [2,2], 3).
A = array(array(_, _, _), array(_, 3, _)).

?- array(A, [2, 3]), cell(A, [2], X).
A = array(array(_, _, _), array(_A, _B, _C)),
X = array(_A, _B, _C).

?- array_list(A, [2,3,4]), let([I=A[1],V=A[I]], writeln(a(I,V))).
a(2,3)
A = array(2, 3, 4).
  
?- array(A, [2, 3]), (set_cell(A, [1], 9) ; nb_set_cell(A, [2], 5); set_cell(A, [2,2],8)).
A = array(array(9, 9, 9), array(_, _, _)) ;
A = array(array(_, _, _), array(5, 5, 5)) ;
A = array(array(_, _, _), array(5, 8, 5)).
==
  
Array[Indices] functional notation defined here using multifile user:shorthand/3 predicate of library(comprehension)
can be used in "in" and "where" conditions of comprehension metapredicates and in constraints of library(clp).

==
?- array(A, [5]), for_all([I in 1..5], A[I] #= I).
A = array(1, 2, 3, 4, 5).
==

*/

%:- catch(reexport(library(comprehension)), _, (throw(error(pack_comprehension_is_not_installed)), fail)).
:- reexport(library(comprehension)).


				% ARRAY CREATION

%! array(+Term)
% 
% tests whether a given term is an array without checking multidimensional consistency.

array(Term):-
    compound(Term),
    functor(Term, array, _).



%! array(?Array, ?DimensionList)
% 
% Array is an array of dimension DimensionList.
% Either creates an array of given dimensions greater or equal to 1,
% or returns the dimensions of a given array,
% or enumerates single dimension arrays.


array(Array, Dimensions):-
    expand(Array, A),
    expand(Dimensions, D),
    (var(A), (var(D) ; D=[N], var(N))
    ->
     D=[N],
     between(1, inf, N),
     functor(A, array, N)
    ;
     array_rec(A, D)).

array_rec(Array, Dimensions):-
    (compound(Array)
    ->
     functor(Array, array, N),
     Dimensions=[N|Tail],
     for_all(I in 1..N,
	     let([Row=Array[I]],
		 (arrays:array_rec(Row, Tail) -> true ; Tail=[])))
    ;
     nonvar(Dimensions),
     Dimensions = [N | Tail],
     nonvar(N),
     functor(Array, array, N),
     (Tail=[]
     ->
      true
     ;
      for_all(I in 1..N, let([Row=Array[I]],
			     (arrays:array_rec(Row, Tail)) -> true; Tail=[])))).


				% CELL ACCESS
    
    
%! cell(+Array, +Indices, ?Cell)
% 
% Cell is the Array cell at given Indices (list of indices for a multidimensional array).
% Throws an error if the indices are out of range.
% Shorthand expressions in Indices are evaluated.

cell(Array, I, _Term):-
    must_be(compound, Array),
    must_be(nonvar, I),
    fail.

cell(Array, [Ind | Indices], Term):-
    !,
    evaluate(Ind, I),
    cell(Array, I, Row),
    (Indices=[]
    ->
     Term=Row
    ;
     cell(Row, Indices, Term)).

cell(Array, Expr, Term):-
    evaluate(Expr, I),
    arg(I, Array, Term).



%! cell(+ArrayIndices, ?Cell)
% 
% Just a shorthand for cell(Array, Indices, Cell) for ArrayIndices of the shorthand/3 form Array[Indices]
cell(ArrayIndices, Cell):-
    must_be(nonvar, ArrayIndices),
    ArrayIndices=Array[Indices],
    cell(Array, Indices, Cell).


				% SHORTHAND FUNCTIONAL NOTATION FOR ARRAY ACCESS


:- multifile user:shorthand/3.

%! shorthand(+Term, +Expanded, +Goal)
%
% Multifile predicate defined here for array functional notation Array[Indices]

user:shorthand(Array[Indices], Var, cell(Array, Indices, Var)):- !.



				% CONVERSIONS BETWEEN ARRAYS AND LISTS

%! array_list(?Array, ?List)
% 
% List is the flat list of the array cells with lexicographically ordered indices.
% Either creates the List or a one dimensional Array indexed by integers starting from 1.
% For a one dimensional array, there is no difference with array_lists/2.

array_list(Array, List):-
    (
     array(Array)
    ->
     array_to_lists(Array, Lists),
     flatten(Lists, List)
    ;
     Array =.. [array | List]
    ).


%! array_lists(+Array, ?List)
% 
% List is the list (of lists in the case of a multidimensional array) of the array cells with lexicographically ordered indices.
% Either creates the lists or the array indexed by integers in intervals starting from 1.
% For a one dimensional array, there is no difference with array_list/2.

array_lists(Array, Lists):-
    (
     array(Array)
    ->
     array_to_lists(Array, Lists)
    ;
     lists_to_array(Lists, Array)
    ).


array_to_lists(Array, List):-
    Array =.. [array | Rows],
    (
     (Rows=[R | _], array(R))
    ->
     call_list(arrays:array_to_lists, Rows, List)
    ;
     List=Rows
    ).

lists_to_array(Lists, Array):-
    must_be(list, Lists),
    length(Lists, N),
    array(Array, [N]),
    for_all(I in 1..N,
	    exists([AI, LI],
		   (cell(Array, I, AI), nth1(I, Lists, LI), (is_list(LI) -> arrays:lists_to_array(LI, AI) ; LI=AI))
		  )
	   ).





				% IMPERATIVE CELL ASSIGNMENT


%! set_cell(+Array, +Indices, ?Term)
% 
% backtrackable assignment of Term to either simple array cell or all subarray cells at given indices.
% Shorthand expressions in Indices are evaluated.

set_cell(Array, I, _Term):-
    must_be(compound, Array),
    must_be(nonvar, I),
    fail.

set_cell(Array, [I | Indices], Term):-
    !,
    (Indices=[]
    ->
     set_all_cells(Array, I, Term)
    ;
     cell(Array, I, Row),
     set_cell(Row, Indices, Term)).

set_cell(Array, Expr, Term):-
    evaluate(Expr, I),
    set_all_cells(Array, I, Term).


%! set_cell(+ArrayIndices, ?Cell)
% 
% Just a shorthand for set_cell(Array, Indices, Cell) for ArrayIndices of the shorthand/3 form Array[Indices]
set_cell(ArrayIndices, Cell):-
    must_be(nonvar, ArrayIndices),
    ArrayIndices=Array[Indices],
    set_cell(Array, Indices, Cell).


set_all_cells(Array, I, Term):-
    arg(I, Array, A),
    (compound(A),
     functor(A, array, N)
    ->
     for_all([J in 1..N], arrays:set_all_cells(A, J, Term))
    ;
     setarg(I, Array, Term)
    ).


%! nb_set_cell(+Array, +Indices, ?Term)
% 
% backtrackable assignment of Term to either simple array cell or all subarray cells at given indices.
% Shorthand expressions in Indices are evaluated.

nb_set_cell(Array, I, _Term):-
    must_be(compound, Array),
    must_be(nonvar, I),
    fail.

nb_set_cell(Array, [I | Indices], Term):-
    !,
    (Indices=[]
    ->
     nb_set_all_cells(Array, I, Term)
    ;
     cell(Array, I, Row),
     nb_set_cell(Row, Indices, Term)).

nb_set_cell(Array, Expr, Term):-
    evaluate(Expr, I),
    nb_set_all_cells(Array, I, Term).


%! nb_set_cell(+ArrayIndices, ?Cell)
% 
% Just a shorthand for nb_set_cell(Array, Indices, Cell) for ArrayIndices of the shorthand/3 form Array[Indices]
nb_set_cell(ArrayIndices, Cell):-
    must_be(nonvar, ArrayIndices),
    ArrayIndices=Array[Indices],
    nb_set_cell(Array, Indices, Cell).



nb_set_all_cells(Array, I, Term):-
    arg(I, Array, A),
    (compound(A),
     functor(A, array, N)
    ->
     for_all([J in 1..N], arrays:nb_set_all_cells(A, J, Term))
    ;
     nb_setarg(I, Array, Term)
    ).

