/*
  comprehension.pl
  
@author Francois Fages
@email Francois.Fages@inria.fr
@license LGPL-2
@version 1.1.3 

  List comprehension for constraint quantifiers, aggregates, and let bindings.
  
*/


:- module(
	  comprehension,
	  [
	   for_all/2,
	   list_of/3,
	   aggregate_for/6,

	   exists/2,
	   let/2,
	   
	   apply_list/2,
	   call_list/2,
	   call_list/3,
	   call_list/4,
	   call_list/5,
	   call_list/6,
	   
	   op(501, xfx, ..),
	   op(502, yfx, \/),
	   
	   op(700, xfx, in),
	   op(990, xfx, where)
	  ]).

/** <module> List comprehension for constraint quantifiers, aggregates, and let bindings.

@author Francois Fages
@version 1.1.3

Prolog metapredicates for set comprehension like setof/3 were introduced by David Warren in 1982
to collect goal instantiations by backtracking.
This necessitated indeed a special mechanism to add to Prolog,
which is used for bagof/3, findall/3, forall/2, foreach/2, or aggregate/3 for instance.
  
This approach to set comprehension is correct for the success set semantics of Prolog,
but incorrect for the constraint semantics of Prolog.

E.g. universal quantifier constraint for_all/2 defined here, compared to standard forall/2 foreach/2:
==
?- L=[A, B, C], for_all(X in L, X=1). % true for all 
L = [1, 1, 1],
A = B, B = C, C = 1.

?- L=[A, B, C], forall(member(X, L), X=1). % satisfiable for all, independently
L = [A, B, C].

?- L=[A, B, C], foreach(member(X, L), X=1). % idem
L = [A, B, C].

?- L=[A, B, C], for_all([X, Y] in L, X=Y).
L = [C, C, C],
A = B, B = C.

?- L=[A, B, C], forall((member(X, L), member(Y, L)), X=Y).
L = [A, B, C].
==

We provide here metapredicates for comprehension by iteration, instead of backtracking, which
  - rename comprehension variables and raise an error if they are instantiated (they can be attributed and will be renamed without attributes),
  - constrain the calling context variables,
  - possibly non-deterministically, if the conditions or goal are non-deterministic.

The comprehension variables must be given with a mandatory "in" domain, and optionally a "where" condition.
  
The "in" domain must be either a domain of integers (i.e. union \/ of intervals ..) or a list of terms.
Shorthand notations are evaluated in the domain.
==
?- for_all([I, J] in 1..3 where I<J, writeln(I-J)).
1-2
1-3
2-3
true.

?- for_all([I in 1..3, J in I+1..3], writeln(I-J)).
1-2
1-3
2-3
true.
==

list_of/3 collects the elements defined by comprehension in a list:
==
?- list_of([I, J] in 1..3 where I<J, I-J, List).
List = [1-2, 1-3, 2-3].
==

aggregate_for/6 expresses a relation on the application of a binary operator between elements defined by comprehension:
  
==
?- aggregate_for(X in 1..5, X+1, +, 0, =, Term).
Term = 1+1+(2+1+(3+1+(4+1+(5+1+0)))).

?- use_module(library(clpfd)).
true.
  
?- aggregate_for(X in 1..5, X+1, +, 0, #=, Result).
Result = 20.
==


The "where" condition in intensional definitions can bind or constrain context variables.
In particular, if the "where" condition is a constraint, the constraint is posted and thus tested for satisfiability, not entailment.
The example below illustrates the difference: X=Y is satisfiable and posted for all X, Y,
whereas in the second query, the condition X==Y is just a test and succeeds on the elements where it is satisfied without changing L:

==
?- L=[A, B, C], for_all([X, Y] in L where X=Y, true).
L = [C, C, C],
A = B, B = C.

?- L=[A, B, C], for_all([X, Y] in L where X==Y, true).
L = [A, B, C].

?- use_module(library(clpfd)).
true.

?- L=[A, B, C], forall(member(X, L), X#>1).
L = [A, B, C].

?- L=[A, B, C], for_all(X in L, X#>1).
L = [A, B, C],
clpfd:(A in 2..sup),
clpfd:(B in 2..sup),
clpfd:(C in 2..sup).

?- forall(member(X, [1, 2, 3]), Y#>X).
true.

?- for_all(X in [1, 2, 3], Y#>X).
clpfd:(Y in 4..sup).
==
  
The "where" condition goal can also be non-deterministic,
making the comprehension metapredicate non-deterministic in this case.

*/

:- reexport(shorthand).


				% BOUNDED UNIVERSAL QUANTIFIER


%! for_all(+VarDomains, +Goal)		  
%
% universal quantifier for iteratively executing Goal on all instances defined in comprehension
% with mandatory "in" domain and optional "where" condition, both allowing user:shorthand/3 functional notations.
% An error is raised a universally quantified variable to be renamed is instantiated.
% The "in" domain must be either a domain of integers (i.e. union of intervals) or a list of terms.
% 
% Unlike forall/2, foreach/2, the conjunction of Goal instances is executed to constraint context variables:
% ==
% ?- L=[A, B, C], for_all(X in L, X=1).
% L = [1, 1, 1],
% A = B, B = C, C = 1.
% ==
% The "where" condition goal can be non-deterministic.
%
% for_all/2 is similar to maplist/2 but using a Goal pattern with arguments defined by comprehension.
    
for_all(VarDomains, Goal) :-
    var_domains(VarDomains, VarDoms, Condition),
    variables(VarDoms, Vars),
    copy_term_nat(Vars, f(VarDoms, Condition, Goal), _, f(VDoms, Cond, G)), % needed in VarDoms as X in [Y] unifies X=Y
    call_comprehension(VDoms, [], [], Cond, G).



% generator comprehension

var_domains(VarDomains, _VarDoms, _Condition) :-
    var(VarDomains),
    !,
    must_be(nonvar,VarDomains). 

var_domains(Var where Condition, [Var], Condition) :-
    var(Var),
    !.

var_domains(Vars in Domain where Condition, VarDoms, Condition) :-
    !,
    (var(Vars)
    ->
     VarDoms=[Vars in Domain]
    ;
     distribute(Vars, Domain, VarDoms)).

var_domains(Vars in Domain, VarDoms, Condition) :-
    !,
    var_domains(Vars in Domain where true, VarDoms, Condition).

var_domains(VarDomains where Condition, VarDomains, Condition) :-
    !.

var_domains(VarDomains, VarDoms, Condition) :-
    must_be(list, VarDomains),
    var_domains(VarDomains where true, VarDoms, Condition).


variables([], []).
variables([VarDom | VarDoms], [V | Vars]):-
    (var(VarDom)
    ->
     V=VarDom
    ;
     VarDom = (V in _)
    ->
     must_be(var, V)
    ;
     must_be(var, VarDom) % instantiated variable 
    ),
    variables(VarDoms, Vars).


% allow only variables to distribute domain to.
distribute([], _, []).
distribute([Var | Vars], Domain, [Var in Domain | Domains]):-
    must_be(var, Var),
    distribute(Vars, Domain, Domains).


    
% iteratively calls Goal on valuations of Vars in Domains satisfying a possibly non-deterministic where Condition
% Vars and Values are built in reversed order
call_comprehension([], Vars, Values, Condition, Goal) :-
    copy_term_nat(Vars, Condition, Values, Cond),
    (expand(Cond)		% one great use of soft cut at last !
    *->
     copy_term_nat(Vars, Goal, Values, G),
     G
    ;
     true).

call_comprehension([VarD | VarDomains], Vars, Values, Condition, Goal) :-
    must_be(nonvar, VarD),
    VarD = (V in VarDomain),
    expand(VarDomain, VarDom),
    (next_value(VarDom, Min, Greater)
    ->
     Vs = [V | Vars],
     Vals = [Min | Values],
     copy_term_nat([V], VarDomains, [Min], VarDoms), % partial evaluation
     call_comprehension(VarDoms, Vs, Vals, Condition, Goal), % to append to next valuations
     (Greater=[]
     ->
      true
     ;
      call_comprehension([V in Greater | VarDomains], Vars, Values, Condition, Goal)
     )
    ;
     true).


next_value(Min .. Max, Mi, Greater):-
    Mi is Min, % shorthand expansions of domains are already done
    Ma is Max,
    (Mi < Ma
    ->
     Mi1 is Mi+1,
     Greater = Mi1..Ma
    ;
     Mi = Ma, % otherwise fail, empty domain
     Greater = []).

next_value(Domain1 \/ Domain2, Min, Greater):-
    nextvalue(Domain1, Min, Great),
    (Great = []
    ->
     Greater = Domain2
    ;
     Greater = Great \/ Domain2).

next_value([Min | List], Mi, List):-
    catch(evaluate(Min, Mi), _, Mi=Min). % just to allow non numerical values in a domain list



				% LIST OF VALUATIONS DEFINED BY IN AND WHERE CONDITIONS

%! list_of(+VarDomains, +Term, ?List)
%
% List is the list of instances of Term
% defined in comprehension with mandatory "in" domain and optional "where" condition both allowing user:shorthand/3 functional notations.
%
% ``list_of(VarDomains, Expr, List)`` is defined by ``aggregate_for(VarDomains, Expr, [|], [], =, List)``.
%
% An error is raised if the comprehension variables to be renamed are instantiated
% (they can be attributed and will be renamed without attributes).
%
% The "where" condition goal can be non-deterministic.

list_of(VarDomains, Expr, List):-
    aggregate_for(VarDomains, Expr, '[|]', [], =, List).



				% GENERAL AGGREGATE_FOR PREDICATE

%! aggregate_for(+VarDomains, +Expression, +Op, +E, +Relation, ?Term)
%
% true iff Relation(Term, Aggregate) holds where Aggregate is the application of binary operation Op with starting element E,
% to all instances of Expression defined in comprehension by VarDomains with mandatory "in" domain and optional "where" condition.
% An error is raised if a comprehension variable is instantiated (if it is attributed it is renamed without attributes).
%
% For example, list_of(VarDomains, Expression, List) is defined here by ``aggregate_for(VarDomains, Expression, [|], [], =, List)``.
%
% The instances of Expression are not evaluated unless Expression is of the form "Variable where Goal" in which case the Goal is executed.
% It is an error if the first argument of where is not a variable. The variable is existentially quantified and renamed in the Goal.
% 
% The list of arguments of a term could be defined by
% ==
% ?- Term=f(a,g(X),Y), functor(Term, _, N),  aggregate_for([I in 1..N], Arg where arg(I, Term, Arg), '[|]', [], =, Args).
% Term = f(a, g(X), Y),
% N = 3,
% Args = [a, g(X), Y].
% ==
%
% aggregate_for/6 is similar to foldl/4 but more general and using Goal pattern with arguments defined in comprehension.
%
% Used in library(clp) to define some hybrid global constraints.

aggregate_for(VarDomains, Expression, Op, E, Relation, Term):-
    (compound(Expression), Expression=(X where _Goal)
    ->
     must_be(var, X),
     copy_term_nat([X], Expression, _, Expr)
    ;
     Expr=Expression),
    var_domains(VarDomains, VarDoms, Condition),
    variables(VarDoms, Vars),
    copy_term_nat(Vars, f(VarDoms, Condition, Expr), Vs, f(VDoms, Cond, Exp)), % needed in VarDoms since X in [Y] unifies X=Y
    values(VDoms, [], [], Cond, ValuesList),
    aggregate_rec(ValuesList, Vs, Exp, Op, E, Aggregate),
    Rel =.. [Relation, Term, Aggregate],
    expand(Rel).

aggregate_rec([], _, _, _, E, E).
aggregate_rec([Values | ValuesList], Vars, Expression, Op, E, Term) :-
    copy_term_nat(Vars, Expression, Values, Expr),
    (compound(Expr), Expr=(X where _Goal)
    ->
     must_be(var, X),
     copy_term_nat([X], Expr, _, Exp where Goal),
     expand(Goal) % binding or constraining X in Expr renamed
    ;
     expand(Expr, Exp)), %Exp=Expr),
    Term =.. [Op, Exp, Aggregate], % TODO do not construct the full aggregate_for expression
    aggregate_rec(ValuesList, Vars, Expression, Op, E, Aggregate).


% computes the List of valuations of Vars in Domains satisfying a, possibly non-deterministic, where Condition
values([], Vars, Values, Condition, List):-
    copy_term_nat(Vars, Condition, Values, Cond),
    (expand(Cond)				% one great use of soft cut at last !
    *->
     reverse(Values, Vals),
     List=[Vals]
    ;
     List = []).

values([VarD | VarDomains], Vars, Values, Condition, List):-
    must_be(nonvar, VarD),
    VarD = (V in VarDomain),
    expand(VarDomain, VarDom),
    (next_value(VarDom, Min, Greater)
    ->
     Vs = [V | Vars],
     Vals = [Min | Values],
     copy_term_nat([V], VarDomains, [Min], VarDoms), % partial evaluation
     values(VarDoms, Vs, Vals, Condition, List1), % to append to next valuations
     (Greater=[]
     ->
      List = List1
     ;
      values([V in Greater | VarDomains], Vars, Values, Condition, List2),
      append(List1, List2, List))
    ;
     List = []).



				% EXISTENTIAL QUANTIFIER AND LET EXPRESSIONS

%! exists(+Vars, +Goal)
%
% calls Goal with existentially quantified variables Vars.
% An error is raised if the existential variables to rename are instantiated
% (they can be attributed though and will be renamed without attributes).

exists(Vars, Goal):-
    must_be(list(var), Vars), 
    copy_term_nat(Vars, Goal, _, RenamedGoal),
    RenamedGoal.



%! let(+Bindings, +Goal)
%
% calls Goal with Bindings for existentially quantified variables bound to expressions allowing user:shorthand/3 functional notations.
% An error is raised if the existential variables to rename are instantiated
% (they can be attributed and will be renamed without attributes).

let(Bindings, Goal):-
    must_be(list, Bindings),
    var_bindings(Bindings, Vars),
    must_be(list(var), Vars), 
    copy_term_nat(Vars, f(Bindings, Goal), _, f(BindingsRenamed, GoalRenamed)),
    bindings(BindingsRenamed),
    GoalRenamed.


:- multifile user:shorthand/3.

%! user:shorthand(+Term, +Expanded, +Goal)
%
% Multifile user:shorthand predicate from library(shorthand)
%
% Defined here for expanding let/2 inside expressions, by
% 
%     user:shorthand(let(Bindings, Expr), RenamedExpr, let(Bindings, RenamedExpr=Expr)) :- !.
%
% ==
% ?- evaluate(2*let([X=3], let([Y=2], X+Y)), R).
% R = 10.
% ==

user:shorthand(let(Bindings, Expr), Ex, (expand(Expr, Exp), let(Bindings, Ex=Exp))) :- !.



% constrains the let variables

bindings([]).
bindings([Binding | Bindings]):-
    (var(Binding)
    ->
     % allows using let/2 for unbound existentially quantified variables as in exists/2
     bindings(Bindings)
    ;
     Binding =.. [Binder, X, Term],
     must_be(var, X),
     Goal =.. [Binder, X, Term],
     (member(Binder, [=, is, =..])
     ->
      expand(Goal)
     ;
      member(Binder, [in, ins, #=, #<==>, #<, #>, #=<, #>=, #\=]) % whatever makes sense for a let
     ->
      Goal
     ;
      throw(error(unauthorized_let_binder(Binding)))),
     bindings(Bindings)).

var_bindings([], []).
var_bindings([Binding | Bindings], [X |Vars]) :-
    (var(Binding)
    ->
     X=Binding
    ;
     compound(Binding), Binding =.. [_, X, _]
    ->
     must_be(var, X)
    ;
     throw(error(unauthorized_let_expression(Binding)))),
    var_bindings(Bindings, Vars).


				% CONJUNCTIVE APPLICATION OF A GOAL TO ARGUMENTS IN LISTS

:- meta_predicate apply_list(1, ?).

%! apply_list(+Goal, +ArgsList)
%
% Conjunctive application of Goal to list of list of arguments ArgsList. Fails if the conjunction is not satisfiable.
% Similar to maplist on transposed lists of arguments.

apply_list(Goal, ArgsList):-
    list_apply(ArgsList, Goal).


list_apply([], _).
list_apply([Args | ArgsList], Goal):-
    apply(Goal, Args),
    list_apply(ArgsList, Goal).


:- meta_predicate call_list(1, ?).
:- meta_predicate call_list(2, ?, ?).
:- meta_predicate call_list(3, ?, ?, ?).

:- meta_predicate list_call(?, 1).
:- meta_predicate list_call(?, ?, 2).
:- meta_predicate list_call(?, ?, ?, 3).

%! call_list(+Goal, ?Args)
%
% More efficient conjunctive application of a unary Goal to each argument in list Args.
% Similar to maplist on transposed lists of arguments.

call_list(Goal, Args1):-
    list_call(Args1, Goal).

list_call([], _).
list_call([Arg1 | Args1], Goal):-
    call(Goal, Arg1),
    list_call(Args1, Goal).


%! call_list(+Goal, ?Args1, ?Args2)
%
% Conjunctive application of a binary Goal to the arguments in lists Args1, Args2.
% Similar to maplist on transposed lists of arguments.

call_list(Goal, Args1, Args2):-
    list_call(Args1, Args2, Goal).

list_call([], [], _).
list_call([Arg1 | Args1], [Arg2 | Args2], Goal):-
    call(Goal, Arg1, Arg2),
    list_call(Args1, Args2, Goal).


%! call_list(+Goal, ?Args1, ?Args2, ?Args3)
%
% Conjunctive application of a ternary Goal to the arguments in lists Args1, Args2, Args3.
% Similar to maplist on transposed lists of arguments.

call_list(Goal, Args1, Args2, Args3):-
    list_call(Args1, Args2, Args3, Goal).

list_call([], [], [], _).
list_call([Arg1 | Args1], [Arg2 | Args2], [Arg3 | Args3], Goal):-
    call(Goal, Arg1, Arg2, Arg3),
    list_call(Args1, Args2, Args3, Goal).


%! call_list(+Goal, ?Args1, ?Args2, ?Args3, ?Args4)
%
% Conjunctive application of a 4-ary Goal.
% Similar to maplist on transposed lists of arguments.

call_list(Goal, Args1, Args2, Args3, Args4):-
    list_call(Args1, Args2, Args3, Args4, Goal).

list_call([], [], [], [], _).
list_call([Arg1 | Args1], [Arg2 | Args2], [Arg3 | Args3], [Arg4 | Args4], Goal):-
    call(Goal, Arg1, Arg2, Arg3, Arg4),
    list_call(Args1, Args2, Args3, Args4, Goal).


%! call_list(+Goal, ?Args1, ?Args2, ?Args3, ?Args4, ?Args5)
%
% Conjunctive application of a 5-ary Goal.
% Similar to maplist on transposed lists of arguments.

call_list(Goal, Args1, Args2, Args3, Args4, Args5):-
    list_call(Args1, Args2, Args3, Args4, Args5, Goal).

list_call([], [], [], [], [], _).
list_call([Arg1 | Args1], [Arg2 | Args2], [Arg3 | Args3], [Arg4 | Args4], [Arg5 | Args5], Goal):-
    call(Goal, Arg1, Arg2, Arg3, Arg4, Arg5),
    list_call(Args1, Args2, Args3, Args4, Args5, Goal).


