/*
  shorthand.pl
  
@author Francois Fages
@email Francois.Fages@inria.fr
@license LGPL-2
@version 1.1.3 

  General purpose definitions of shorthand functional notations with run-time expansion.
  
*/


:- module(
	  shorthand,
	  [
	   expand/1,
	   expand/2,
	   expand_subterms/2,
	   evaluate/2
	  ]).

/** <module> General purpose definitions of shorthand functional notations with run-time expansion.

@author Francois Fages
@version 1.1.3


This library provides general purpose multifile user:shorthand/3 and expand/1 expand/2 metapredicates for introducing shorthand functional notations.
  - defined here for allowing conditional expression if(Condition, Expr1, Expr2) in expressions;
  - proposed for allowing global variables in expressions by adding clauses to user:shorthand/3;
  - defined in library(comprehension) for let(Bindings, Expr) expressions similarly to let/2 goal predicate defined there;
  - in library(arrays) for Array[Indices] functional notation,
  - in library(clp) for sum(VarDomains, Expr) expressions with sum/3 constraint, and for all other constraints where the last argument can be seen as a result.

==
?- X = -2, expand(Y is if(X>0, X, -X)).
X = -2,
Y = 2.

?- nb_setval(n, 42).
true.

?- asserta(user:shorthand(n, V, nb_getval(n, V))).
true.

?- expand(X is n^2).
X = 1764.
==    
*/


:- multifile user:shorthand/3.
:- dynamic user:shorthand/3. % just for adding shorthand clauses at toplevel as in the example above



%! user:shorthand(+Term, +Expanded, +Goal)
%
% Multifile user:shorthand predicate to define Term as a shorthand for Expanded expression
% with Goal executed at run time (unlike compile time macros).
% 
% Shorthand functional notations are expanded by expand/2, expand/1 and evaluate/2.
% The Goal is not recursively expansed so write ``expand(Goal)`` in the shorthand if it is what you wish.
% 
% The shorthands and the Goal can be non-deterministic, in which case the expansion is non-deterministic.
% 
% One shorthand notation is defined here for conditional expression if(Cond, Expr1, Expr2) by
% 
%     user:shorthand(if(Condition, Expr1, Expr2), Var, expand(Condition -> Var=Expr1 ; Var=Expr2))) :- !.
% 
% The expansion of some global variables can be specified by the user adding shorthand/3 clauses as follows
% (taking care of properly preserving the indexing of user:shorthand/3 on its first argument)
% 
%     user:shorthand(global_variable_name, Value, nb_getval(global_variable_name, Value)):- !.
%
% In library(arrays) a shorthand clause is defined for Array[Indices] functional notation
%
% In library(comprehension) shorthands are expanded in "in" domain and "where" conditions
% and a shorthand is defined for let expressions using let/2 predicate of that library.
% 
% In principle, any n-ary predicate for which the last argument can be interpreted as a result,
% can be transformed in a shorthand functional notation with n-1 arguments.
% E.g. library(clp) uses shorthands in constraints and domains
% and defines shorthand functional notations for constraints where the last argument can be interpreted as a result,
% e.g. sum with 2 arguments, defined using constraint sum/3, is allowed in expressions and constraints.

user:shorthand(if(Condition, Expr1, Expr2), Var, expand((Condition -> Var=Expr1 ; Var=Expr2))) :- !.


%! expand(+Term, ?Expanded)
%
% First-level expansion of Term and its subterms according to multifile definitions of user:shorthand/3.
% The expansion is first-level and not recursive unless the goal in shorthand/3 clause calls expand/2.

expand(Term, Expr):-
    (var(Term)
    ->
     Expr=Term 
    ;
     (user:shorthand(Term, T, Goal)
     ->
      Goal % not expanded because perhaps too early for the arguments inside
     ;
      T=Term),
     expand_subterms(T, Expr)
    ).


%! expand_subterms(+Term, ?Expanded)
%
% recursive expansion of the strict subterms of Term according to multifile definitions of user:shorthand/3.

expand_subterms(Term, Expanded):-
    (compound(Term)
    ->
     (Term=[T | Tail]
     ->
      Expanded=[ET | ETail],
      expand_subterms(T, ET),
      expand_subterms(Tail,ETail)
     ;
      Term =.. [F | Subterms],
      call_list(expand, Subterms, ExpandedSubterms),
      Expanded =.. [F | ExpandedSubterms]
     )
    ;
     Expanded=Term 
    ).
    

%! expand(+Goal)
%
% expands and calls Goal.

expand(Goal) :-
    expand(Goal, G),
    G.


%! evaluate(+Expr, ?Number)
%
% evaluates an arithmetic expression with user:shorthand/3 functional notations.

evaluate(Expr, Number):-
    expand(Expr, E),
    Number is E.
